/* RoomVmeAccess.cc */
/* Edited by Sanshiro E. on 6 October 1997. */
/* Last updated by Sanshiro E. on 18 September 2000. */


#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include "RoomVmeAccess.hh"


TVmeController::TVmeController(void)
{
}

TVmeController::~TVmeController()
{
}

TVmeAccessProperty* TVmeController::Open(TVmeTransferMode TransferMode) throw(THardwareException)
{
    if (TransferMode > tmNumberOfVmeTransferModes) {
	throw THardwareException(
            "TVmeController::Open()",
            "Bad vme transfer mode"
        );	
    }

    int DeviceDescriptor = open(DeviceFileName(TransferMode), O_RDWR);
    if (DeviceDescriptor < 0) {
	throw THardwareException(
            "TVmeController::Open(): open(2)",
            strerror(errno)
        );
    }

    TVmeAccessProperty* Property = new TVmeAccessProperty;
    Property->TransferMode = TransferMode;
    Property->VmeAddress = 0;
    Property->DeviceDescriptor = DeviceDescriptor;
    Property->DmaDeviceDescriptor = 0;
    Property->MappedAddress = 0;
    Property->MapSize = 0;
    Property->InterruptNumber = 0;
    Property->InterruptVector = 0;
    Property->DeviceSpecificData = 0;

    Seek(Property, 0);
    
    return Property;
}

void TVmeController::Close(TVmeAccessProperty* Property)
{
    DmaClose(Property);
    close(Property->DeviceDescriptor);
    delete Property;
}

long TVmeController::Read(TVmeAccessProperty* Property, void* Buffer, int Size) throw(THardwareException)
{
    long ReadSize = read(Property->DeviceDescriptor, Buffer, Size);

    if (ReadSize < 0) {
	throw THardwareException(
            "TVmeController::Read(): read(2)",
            strerror(errno)
        );
    }

    return ReadSize;
}

long TVmeController::Write(TVmeAccessProperty* Property, const void* Buffer, int Size) throw(THardwareException)
{
    long WrittenSize = write(Property->DeviceDescriptor, Buffer, Size);

    if (WrittenSize < 0) {
	throw THardwareException(
            "TVmeController::Write(): write(2)",
            strerror(errno)
        );
    }

    return WrittenSize;
}

off_t TVmeController::Seek(TVmeAccessProperty* Property, off_t OffsetAddress) throw(THardwareException)
{
    off_t SeekPosition = Property->VmeAddress + OffsetAddress;

    off_t NewPosition;
    NewPosition = lseek(Property->DeviceDescriptor, SeekPosition, SEEK_SET);

    if (NewPosition < 0) {
	throw THardwareException(
            "TVmeController::Seek(): lseek(2)",
            strerror(errno)
        );	
    }

    return NewPosition;
}

caddr_t TVmeController::Map(TVmeAccessProperty* Property, off_t VmeAddress, size_t MapSize) throw(THardwareException)
{
    caddr_t MappedAddress = 0;

    if (MapSize > 0) {
	caddr_t MapAddress = (caddr_t) 0;  // assigned by kernel
	int MapProtection = PROT_READ | PROT_WRITE;
	int MapFlag = MAP_SHARED;
	MappedAddress = (caddr_t) mmap(
	    MapAddress,
	    MapSize,
	    MapProtection,
	    MapFlag,
	    Property->DeviceDescriptor,
	    VmeAddress
	);
	if (MappedAddress == (caddr_t) MAP_FAILED) {
	    throw THardwareException(
		"TVmeController::Map(): mmap(2)",
		strerror(errno)
	    );
	}
    }

    Property->VmeAddress = VmeAddress;
    Property->MapSize = MapSize;
    Property->MappedAddress = MappedAddress;

    return MappedAddress;
}

void TVmeController::Unmap(TVmeAccessProperty* Property) throw(THardwareException)
{
    if (Property->MappedAddress > 0) {
	if (munmap((caddr_t) Property->MappedAddress, Property->MapSize) < 0) {
	    throw THardwareException(
		"TVmeController::Unmap(): munmap(2)",
		strerror(errno)
	    );	     
	}
    }
}

void TVmeController::RegisterInterruptNotification(TVmeAccessProperty* Property, int InterruptNumber, int Vector, int SignalId) throw(THardwareException)
{
    Property->InterruptNumber = InterruptNumber;
    Property->InterruptVector = Vector;
    Property->InterruptSignalId = SignalId;    
}

void TVmeController::DmaOpen(TVmeAccessProperty* Property) throw(THardwareException)
{
    TVmeTransferMode TransferMode = Property->TransferMode;
    int DmaDeviceDescriptor = open(DmaDeviceFileName(TransferMode), O_RDWR);
    if (DmaDeviceDescriptor < 0) {
	throw THardwareException(
            "TVmeController::DmaOpen(): open(2)",
            strerror(errno)
        );
    }

    Property->DmaDeviceDescriptor = DmaDeviceDescriptor;
    DmaSeek(Property, 0);
}

void TVmeController::DmaClose(TVmeAccessProperty* Property)
{
    if (Property->DmaDeviceDescriptor > 0) {
	close(Property->DmaDeviceDescriptor);
    }
}

long TVmeController::DmaRead(TVmeAccessProperty* Property, void* Buffer, int Size) throw(THardwareException)
{
    if (Property->DmaDeviceDescriptor <= 0) {
	DmaOpen(Property);
    }  

    long ReadSize = read(Property->DmaDeviceDescriptor, Buffer, Size);

    if (ReadSize < 0) {
	throw THardwareException(
            "TVmeController::DmaRead(): read(2)",
            strerror(errno)
        );
    }

    return ReadSize;
}

long TVmeController::DmaWrite(TVmeAccessProperty* Property, const void* Buffer, int Size) throw(THardwareException)
{
    if (Property->DmaDeviceDescriptor <= 0) {
       DmaOpen(Property);
    }  

    long WrittenSize = write(Property->DmaDeviceDescriptor, Buffer, Size);

    if (WrittenSize < 0) {
	throw THardwareException(
            "TVmeController::DmaWrite(): write(2)",
            strerror(errno)
        );
    }

    return WrittenSize;
}

off_t TVmeController::DmaSeek(TVmeAccessProperty* Property, off_t OffsetAddress) throw(THardwareException)
{
    if (Property->DmaDeviceDescriptor <= 0) {
       DmaOpen(Property);
    }  

    off_t SeekPosition = Property->VmeAddress + OffsetAddress;

    off_t NewPosition;
    NewPosition = lseek(Property->DmaDeviceDescriptor, SeekPosition, SEEK_SET);

    if (NewPosition < 0) {
	throw THardwareException(
            "TVmeController::DmaSeek(): lseek(2)",
            strerror(errno)
        );	
    }

    return NewPosition;
}


TVmeCrate::TVmeCrate(TVmeController *VmeController, int NumberOfSlots) 
{
    _VmeController = VmeController;
    _NumberOfSlots = NumberOfSlots;
}

TVmeCrate::TVmeCrate(int NumberOfSlots) 
{
    _VmeController = 0;
    _NumberOfSlots = NumberOfSlots;
}

TVmeCrate::~TVmeCrate() 
{
}

void TVmeCrate::InstallController(TVmeController *VmeController)
{
    _VmeController = VmeController;
}

void TVmeCrate::Install(TVmeModule *VmeModule, off_t VmeAddress, int InterruptNumber, int InterruptVector, int InterruptSignalId) throw(THardwareException)
{
    if (_VmeController == 0) {
        throw THardwareException(
            "TVmeCrate::Install()",
            "controller is not installed"
        );
    }

    TVmeTransferMode TransferMode = VmeModule->TransferMode();
    size_t MapSize = VmeModule->MapSize();

    TVmeAccessProperty* Property = _VmeController->Open(TransferMode);
    caddr_t MappedAddress = _VmeController->Map(Property, VmeAddress, MapSize);
    
    if (InterruptNumber > 0) {
	_VmeController->RegisterInterruptNotification(
	    Property, InterruptNumber, InterruptVector, InterruptSignalId
	);
    }

    VmeModule->Attach(this, Property, MappedAddress);
    _ModuleList.push_back(VmeModule);
     
    if (InterruptNumber > 0) {
 	VmeModule->EnableInterrupt();
    }
}

void TVmeCrate::Install(TVmeModule *VmeModule, off_t VmeAddress) throw(THardwareException)
{
    int InterruptNumber = 0;
    int Vector = 0;

    Install(VmeModule, VmeAddress, InterruptNumber, Vector);
}

void TVmeCrate::Uninstall(TVmeModule *VmeModule) 
{
    VmeModule->DisableInterrupt();
    _VmeController->Unmap(VmeModule->VmeAccessProperty());
    _VmeController->Close(VmeModule->VmeAccessProperty());

    VmeModule->Detach();
}

TVmeController* TVmeCrate::VmeController(void) const
{
    return _VmeController;
}


TVmeModule::TVmeModule(size_t MapSize, TVmeTransferMode TransferMode)
{
    _MapSize = MapSize;
    _TransferMode = TransferMode;

    _VmeController = 0;
    _VmeCrate = 0;
    _VmeAccessProperty = 0;
}

TVmeModule::~TVmeModule()
{
    if (_VmeCrate != 0) {
	_VmeCrate->Uninstall(this);
    }
}

TVmeModule* TVmeModule::Clone(void)
{
    return new TVmeModule(_MapSize, _TransferMode);
}

size_t TVmeModule::MapSize(void) const
{
    return _MapSize;
}

TVmeTransferMode TVmeModule::TransferMode(void) const
{
    return _TransferMode;
}

caddr_t TVmeModule::MappedAddress(void) const
{
    return _MappedAddress;
}

TVmeAccessProperty* TVmeModule::VmeAccessProperty(void) const
{
    return _VmeAccessProperty;
}

long TVmeModule::PioRead(off_t Address, void* Buffer, int Size) throw(THardwareException)
{
    _VmeController->Seek(_VmeAccessProperty, Address);
    return _VmeController->Read(_VmeAccessProperty, Buffer, Size);
}

long TVmeModule::PioWrite(off_t Address, const void* Buffer, int Size) throw(THardwareException)
{
    _VmeController->Seek(_VmeAccessProperty, Address);
    return _VmeController->Write(_VmeAccessProperty, Buffer, Size);
}

long TVmeModule::DmaRead(off_t Address, void* Buffer, int Size) throw(THardwareException)
{
    _VmeController->DmaSeek(_VmeAccessProperty, Address);
    return _VmeController->DmaRead(_VmeAccessProperty, Buffer, Size);
}

long TVmeModule::DmaWrite(off_t Address, const void* Buffer, int Size) throw(THardwareException)
{
    _VmeController->DmaSeek(_VmeAccessProperty, Address);
    return _VmeController->DmaWrite(_VmeAccessProperty, Buffer, Size);
}

int TVmeModule::EnableInterrupt(void) throw(THardwareException)
{
    _VmeController->EnableInterruptNotification(_VmeAccessProperty);

    return 0;
}

int TVmeModule::DisableInterrupt(void) throw(THardwareException)
{
    _VmeController->DisableInterruptNotification(_VmeAccessProperty);

    return 0;
}

int TVmeModule::WaitData(unsigned TimeOut) throw(THardwareException)
{
    return _VmeController->WaitForInterruptNotification( 
	_VmeAccessProperty, TimeOut 
    );
}

  
void TVmeModule::Attach(TVmeCrate* Crate, TVmeAccessProperty* Property, caddr_t MappedAddress)
{
    _VmeCrate = Crate;
    _VmeController = Crate->VmeController();
    _VmeAccessProperty = Property;
    _MappedAddress = MappedAddress;
}

void TVmeModule::Detach(void)
{
    _VmeCrate = 0;
    _VmeController = 0;
    _VmeAccessProperty = 0;
    _MappedAddress = 0;
}

#if 0
int TVmeModule::WaitWord(off_t OffsetAddress, Word Value, unsigned TimeLimit)
{
    unsigned Time = 0;
    int EventOccured;

    while (1) {
	EventOccured = (GetWord(OffsetAddress) == Value);
	if ((EventOccured) || (Time > TimeLimit)) {
	    break;
	}
	
	if (TimeLimit > 0) {
	    Time++;  // sec
	}
	Suspend(1, 0);
    }

    return EventOccured;
}
#endif
