/* RoomVmeAccess.hh */
/* Edited by Sanshiro E. on 6 October 1997. */
/* Last updated by Sanshiro E. on 18 September 2000.*/


#ifndef __RoomVmeAccess_hh__
#define __RoomVmeAccess_hh__

#include <vector>
#include <sys/types.h>
#include "RoomDefs.hh"
#include "RoomDeviceBase.hh"
#include "RoomModule.hh"


enum TVmeTransferMode {
    tmA16D16, tmA16D32, 
    tmA24D16, tmA24D32, 
    tmA32D16, tmA32D32,
    tmNumberOfVmeTransferModes
};


class TVmeAccessProperty {
  public:
    off_t VmeAddress;
    TVmeTransferMode TransferMode;
    int DeviceDescriptor;
    caddr_t MappedAddress;
    size_t MapSize;
    int DmaDeviceDescriptor;
    int InterruptNumber;
    int InterruptVector;
    int InterruptSignalId;
    void* DeviceSpecificData;
};

class TVmeModule;


class TVmeController: public TController {
  public:
    TVmeController(void);
    virtual ~TVmeController();
    virtual TVmeController* Clone(void) = 0;
  public:
    virtual TVmeAccessProperty* Open(TVmeTransferMode TransferMode) throw(THardwareException);
    virtual void Close(TVmeAccessProperty* Property);
    virtual long Read(TVmeAccessProperty* Property, void* Buffer, int Size) throw(THardwareException);
    virtual long Write(TVmeAccessProperty* Property, const void* Buffer, int Size) throw(THardwareException);
    virtual off_t Seek(TVmeAccessProperty* Property, off_t OffsetAddress) throw(THardwareException);
    virtual caddr_t Map(TVmeAccessProperty* Property, off_t VmeAddress, size_t MapSize) throw(THardwareException);
    virtual void Unmap(TVmeAccessProperty* Property) throw(THardwareException);
    virtual void RegisterInterruptNotification(TVmeAccessProperty* Property, int InterruptNumber, int Vector, int SignalId = 0) throw(THardwareException);
    virtual void EnableInterruptNotification(TVmeAccessProperty* Property) throw(THardwareException) = 0;
    virtual void DisableInterruptNotification(TVmeAccessProperty* Property) throw(THardwareException) = 0;
    virtual int WaitForInterruptNotification(TVmeAccessProperty* Property, unsigned TimeOut) throw(THardwareException) = 0;
    virtual long DmaRead(TVmeAccessProperty* Property, void* Buffer, int Size) throw(THardwareException);
    virtual long DmaWrite(TVmeAccessProperty* Property, const void* Buffer, int Size) throw(THardwareException);
    virtual off_t DmaSeek(TVmeAccessProperty* Property, off_t OffsetAddress) throw(THardwareException);
  protected:
    virtual const char* DeviceFileName(TVmeTransferMode TransferMode) = 0;
    virtual const char* DmaDeviceFileName(TVmeTransferMode TransferMode) = 0;
    virtual void DmaOpen(TVmeAccessProperty* Property) throw(THardwareException);
    virtual void DmaClose(TVmeAccessProperty* Property);
};


class TVmeCrate: public TCrate {
  public:
    TVmeCrate(TVmeController *VmeController, int NumberOfSlots = 21);
    TVmeCrate(int NumberOfSlots = 21);
    virtual ~TVmeCrate();
    virtual void InstallController(TVmeController *VmeController);
    virtual void Install(TVmeModule *VmeModule, off_t VmeAddress) throw(THardwareException);
    virtual void Install(TVmeModule *VmeModule, off_t VmeAddress, int InterruptNumber, int InterruptVector, int InterruptSignalId = 0) throw(THardwareException);
    virtual void Uninstall(TVmeModule *VmeModule);
    virtual TVmeController* VmeController(void) const;
  protected:
    TVmeController *_VmeController;
    vector<TVmeModule*> _ModuleList;
    int _NumberOfSlots;
};


class TVmeModule: public TModule {
  public:
    TVmeModule(size_t MapSize, TVmeTransferMode TransferMode);
    virtual ~TVmeModule();
    virtual TVmeModule* Clone(void);
    virtual TVmeTransferMode TransferMode(void) const;
    virtual caddr_t MappedAddress(void) const;
    virtual size_t MapSize(void) const;
    inline Word PutWord(off_t OffsetAddress, Word Value);
    inline Word GetWord(off_t OffsetAddress);
    inline Word* WordPtr(off_t OffsetAddress = 0);
    inline DoubleWord PutDoubleWord(off_t OffsetAddress, DoubleWord Value);
    inline DoubleWord GetDoubleWord(off_t OffsetAddress);
    inline DoubleWord* DoubleWordPtr(off_t OffsetAddress = 0);
    virtual long PioRead(off_t Address, void* Buffer, int Size) throw(THardwareException);
    virtual long PioWrite(off_t Address, const void* Buffer, int Size) throw(THardwareException);
    virtual long DmaRead(off_t Address, void* Buffer, int Size) throw(THardwareException);
    virtual long DmaWrite(off_t Address, const void* Buffer, int Size) throw(THardwareException);
    virtual int EnableInterrupt(void) throw(THardwareException);
    virtual int DisableInterrupt(void) throw(THardwareException);
    virtual int WaitData(unsigned TimeOut) throw(THardwareException);
  private:
    friend class TVmeCrate;
    void Attach(TVmeCrate* Crate, TVmeAccessProperty* Property, caddr_t MappedAddress);
    void Detach(void);
    TVmeAccessProperty* VmeAccessProperty(void) const;
  private:
    TVmeController* _VmeController;
    TVmeCrate* _VmeCrate;
    TVmeAccessProperty* _VmeAccessProperty;
    TVmeTransferMode _TransferMode;
    volatile caddr_t _MappedAddress;
    size_t _MapSize;
    int _InterruptSignalId;
};


inline Word TVmeModule::PutWord(off_t OffsetAddress, Word Value) {
    return *((Word *) (_MappedAddress + OffsetAddress)) = Value;
}

inline Word TVmeModule::GetWord(off_t OffsetAddress) {
    return *((Word *) (_MappedAddress + OffsetAddress));
}

inline Word *TVmeModule::WordPtr(off_t OffsetAddress) {
    return (Word *) (_MappedAddress + OffsetAddress);
}

inline DoubleWord TVmeModule::PutDoubleWord(off_t OffsetAddress, DoubleWord Value) {
    return *((DoubleWord *) (_MappedAddress + OffsetAddress)) = Value;
}

inline DoubleWord TVmeModule::GetDoubleWord(off_t OffsetAddress) {
    return *((DoubleWord *) (_MappedAddress + OffsetAddress));
}

inline DoubleWord *TVmeModule::DoubleWordPtr(off_t OffsetAddress) {
    return (DoubleWord *) (_MappedAddress + OffsetAddress);
}


#endif
