/* module-SoftwareModules.cc */
/* Edited by Sanshiro E. on 23 April 1998. */
/* Last updated by Sanshiro E. on 18 June 2000. */


#include <iostream.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "RoomDeviceFactory.hh"
#include "RoomSoftwareDevice.hh"
#include "module-SoftwareModules.hh"


static TSoftwareModuleCreator Creator1(
    "PoissonGenerator", new TPoissonGenerator()
);
static TSoftwareModuleCreator Creator2(
    "TimerInterrupter", new TTimerInterrupter()
);
static TSoftwareModuleCreator Creator3(
    "SoftwareCounter", new TSoftwareCounter()
);


static void InitializePoisson(void);
static double Poisson(double Lambda);


TPoissonGenerator::TPoissonGenerator(int NumberOfChannels, int NumberOfSamples, long FullScale)
{
    _NumberOfChannels = NumberOfChannels;
    _NumberOfSamples = NumberOfSamples;
    _FullScale = FullScale;

    _ReadoutCount = new int[NumberOfChannels];

    InitializePoisson();
    Clear();
}

TPoissonGenerator::~TPoissonGenerator()
{
    delete _ReadoutCount;
}
    
TSoftwareModule* TPoissonGenerator::Clone(void)
{
    return new TPoissonGenerator(_NumberOfChannels, _NumberOfSamples, _FullScale);
}

int TPoissonGenerator::Clear(int Channel) throw(THardwareException)
{
    if (Channel < 0) {
	for (int Channel = 0; Channel < _NumberOfChannels; Channel++) {
	    _ReadoutCount[Channel] = 0;
	}
    }
    else if (Channel < _NumberOfChannels) {
	_ReadoutCount[Channel] = 0;
    }
    else {
	return -1;
    }

    return 0;
}

int TPoissonGenerator::Read(int Channel, int &Data) throw(THardwareException)
{
    if (_ReadoutCount[Channel] >= _NumberOfSamples) {
	return 0;
    }
    else {
	_ReadoutCount[Channel]++;
    }
    
    double Lambda = 12;
    Data = (int) (Poisson(Lambda) * (_FullScale / (3.0 * Lambda)));

    if (Data > _FullScale) {
	Data = _FullScale;
    }

    return 1;
}

int TPoissonGenerator::NumberOfChannels(void) throw(THardwareException)
{
    return _NumberOfChannels;
}

int TPoissonGenerator::NumberOfSamples(void) throw(THardwareException)
{
    return _NumberOfSamples;
}

static void InitializePoisson(void)
{
    srand48(time(NULL));
}

static double Poisson(double Lambda)
{
    // Poisson distribution
    // P(k) = exp(-Lambda) Lambda^k / k!

    Lambda = exp(Lambda) * drand48();
    int k = 0;
    while (Lambda > 1) {
	Lambda *= drand48();
	k++;
    }

    return k;
}


TSoftwareCounter::TSoftwareCounter(int NumberOfChannels, int NumberOfSamples, long FullScale)
{
    _NumberOfChannels = NumberOfChannels;
    _FullScale = FullScale;

    _ReadoutCount = new int[NumberOfChannels];
    Initialize();
}

TSoftwareCounter::~TSoftwareCounter()
{
    delete _ReadoutCount;
}
    
TSoftwareModule* TSoftwareCounter::Clone(void)
{
    return new TSoftwareCounter(_NumberOfChannels, _FullScale);
}

int TSoftwareCounter::Initialize(int InitialState) throw(THardwareException)
{
    for (int Channel = 0; Channel < _NumberOfChannels; Channel++) {
	_ReadoutCount[Channel] = 0;
    }

    return 0;
}

int TSoftwareCounter::Read(int Channel, int &Data) throw(THardwareException)
{
    Data = _ReadoutCount[Channel];
    
    if (Data >= _FullScale) {
	Data = _FullScale;
    }
    else {
	_ReadoutCount[Channel]++;
    }

    return 1;
}

int TSoftwareCounter::Clear(int Channel) throw(THardwareException)
{
    if (Channel == -1) {
	for (int Channel = 0; Channel < _NumberOfChannels; Channel++) {
	    _ReadoutCount[Channel] = 0;
	}
    }
    else {
	_ReadoutCount[Channel] = 0;
    }

    return 0;
}

int TSoftwareCounter::NumberOfChannels(void) throw(THardwareException)
{
    return _NumberOfChannels;
}

int TSoftwareCounter::NumberOfSamples(void) throw(THardwareException)
{
    return 1;
}


TTimerInterrupter::TTimerInterrupter(long Interval_sec, long Interval_usec)
{
    _Interval_sec = Interval_sec;
    _Interval_usec = Interval_usec;
}

TTimerInterrupter::~TTimerInterrupter()
{
}

TSoftwareModule* TTimerInterrupter::Clone(void)
{
    return new TTimerInterrupter(_Interval_sec, _Interval_usec);
}

int TTimerInterrupter::WaitData(unsigned TimeLimit /* sec */) throw(THardwareException)
{
    static unsigned long TimeToNext_sec = 0;
    static unsigned long TimeToNext_usec = 0;
    if ((TimeToNext_sec + TimeToNext_usec) == 0) {
	TimeToNext_sec = _Interval_sec;
	TimeToNext_usec = _Interval_usec;
    }

    unsigned long Interval_sec;
    unsigned long Interval_usec;
    if ((TimeToNext_sec + TimeToNext_usec / 1.0e6) > TimeLimit) {
	Interval_sec = TimeLimit;
	Interval_usec = 0;
	TimeToNext_sec -= TimeLimit;
    }
    else {
	Interval_sec = TimeToNext_sec;
	Interval_usec = TimeToNext_usec;
	TimeToNext_sec = 0;
	TimeToNext_usec = 0;
    }

    Suspend(Interval_sec, Interval_usec);
    
    return ((TimeToNext_sec + TimeToNext_usec) == 0) ? 1 : 0;
}

int TTimerInterrupter::WriteRegister(int Address, int Data) throw(THardwareException)
{
    switch (Address) {
      case 0:
        _Interval_sec = Data;
        break;
      case 1:
        _Interval_usec = Data;
        break;
      default:
        return -1;
    }

    return 0;
}

int TTimerInterrupter::ReadRegister(int Address, int& Data) throw(THardwareException)
{
    switch (Address) {
      case 0:
        Data = _Interval_sec;
        break;
      case 1:
        Data = _Interval_usec;
        break;
      default:
        return -1;
    }

    return 0;
}
